import numpy as np
from typing import Dict, Tuple
import eagerx
import pybullet
import gym
import copy


class EvaluateEnv(eagerx.BaseEnv):
    def __init__(self, env, engine, episode_timeout, render="pybullet"):
        self.rate = env.rate
        graph = copy.deepcopy(env.graph)
        self._wrapped = env
        if render == "pybullet":
            name = f"{env.name}_pybullet_eval"
            robot = graph.get_spec("quadruped")
            robot.sensors.image.rate = 5
            graph.add_component(robot.sensors.image)
            graph.render(robot.sensors.image, rate=5, encoding="rgb")
            graph.remove("xy_plane")
        else:
            name = f"{env.name}_xyplane_eval"
            xy_plane = graph.get_spec("xy_plane")
            xy_plane.outputs.image.rate = 5

        # Make the backend specification
        from eagerx.backends.single_process import SingleProcess

        backend = SingleProcess.make()

        super(EvaluateEnv, self).__init__(name, self.rate, graph, engine, backend, force_start=True)
        self.timeout_steps = int(episode_timeout * self.rate)
        self.steps = None

    @property
    def observation_space(self) -> gym.spaces.Dict:
        return self._wrapped.observation_space

    @property
    def action_space(self) -> gym.spaces.Dict:
        return self._wrapped.action_space

    def reset(self):
        # Reset number of steps
        self.steps = 0

        # Sample desired states
        states = self.state_space.sample()

        # set camera location
        if "quadruped/image/pos" in states:
            states["quadruped/image/pos"] = np.array([-1, -1, 0.5], dtype="float32")
        if "quadruped/image/orientation" in states:
            tmp = list(pybullet.getQuaternionFromEuler(np.deg2rad([-90, 0, 0])))
            states["quadruped/image/orientation"] = np.array(tmp, dtype="float32")

        # Perform reset
        obs = self._reset(states)
        if "xs_zs" in obs:
            obs["xs_zs"][0][:] = [-0.01354526, -0.26941818, 0.0552178, -0.25434446]
        return obs

    def step(self, action: Dict) -> Tuple[Dict, float, bool, Dict]:
        obs = self._step(action)
        self.steps += 1

        done = self.steps >= self.timeout_steps

        # Determine done flag
        # Set info about episode truncation
        info = {"TimeLimit.truncated": done}
        return obs, 0.0, done, info



import eagerx
from eagerx.core.specs import EngineStateSpec, ObjectSpec
from typing import Any


class DummyState(eagerx.EngineState):
    @classmethod
    def make(cls) -> EngineStateSpec:
        return cls.get_specification()

    def initialize(self, spec: EngineStateSpec, object_spec: ObjectSpec, simulator: Any):
        pass

    def reset(self, state: Any):
        pass


class SetGymAttribute(eagerx.EngineState):
    @classmethod
    def make(cls, attribute: str) -> EngineStateSpec:
        spec = cls.get_specification()
        spec.config.attribute = attribute
        return spec

    def initialize(self, spec: EngineStateSpec, object_spec: ObjectSpec, simulator: Any):
        self.attribute = spec.config.attribute
        self.simulator = simulator

    def reset(self, state: Any):
        for _obj_name, sim in self.simulator.items():
            if hasattr(sim["env"].env, self.attribute):
                setattr(sim["env"].env, self.attribute, state)
            else:
                self.backend.logwarn_once(f"{self.attribute} is not an attribute of the environment.")



from typing import Optional
import numpy as np

# IMPORT EAGERX
from eagerx.core.space import Space
from eagerx.core.constants import process
from eagerx.utils.utils import Msg
from eagerx.core.entities import EngineNode
import eagerx.core.register as register


class FloatOutput(EngineNode):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        idx: Optional[int] = 0,
        process: Optional[int] = process.ENVIRONMENT,
        color: Optional[str] = "cyan",
    ):
        """
        FloatOutput spec

        :param idx: index of the value of interest from the array.
        """
        spec = cls.get_specification()

        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.inputs = ["observation_array"]
        spec.config.outputs = ["observation"]

        # Custom node params
        spec.config.idx = idx
        return spec

    def initialize(self, spec, object_spec, simulator):
        self.obj_name = object_spec.config.name
        self.idx = spec.config.idx

    @register.states()
    def reset(self):
        pass

    @register.inputs(observation_array=Space(dtype="float32"))
    @register.outputs(observation=Space(dtype="float32"))
    def callback(self, t_n: float, observation_array: Optional[Msg] = None):
        data = observation_array.msgs[-1].data
        return dict(observation=np.array(data[self.idx], dtype="float32"))



"""
CPG in polar coordinates based on:
Pattern generators with sensory feedback for the control of quadruped
authors: L. Righetti, A. Ijspeert
https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=4543306
Original author: Guillaume Bellegarda
"""

from typing import Optional
import numpy as np
import eagerx
from eagerx import Space, Node
from eagerx import register
from eagerx.utils.utils import Msg

import eagerx_tutorials.quadruped.go1.configs_go1 as go1_config


class CartesiandPDController(Node):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        process: Optional[int] = eagerx.process.ENVIRONMENT,
    ) -> eagerx.specs.NodeSpec:
        """Make a spec to create a CartesiandPDController node that controls a set of joints.

        :param name: User specified node name.
        :param rate: Rate (Hz) at which the callback is called.
        :param process: Process in which this node is launched. See :class:`~eagerx.core.constants.process` for all options.
        :return: NodeSpec
        """
        spec = cls.get_specification()

        # Modify default node params
        spec.config.update(name=name, rate=rate, process=process)
        spec.config.update(inputs=["cartesian_pos"], outputs=["joint_pos"], joints=list(go1_config.JOINT_NAMES))
        return spec

    def initialize(self, spec):
        self.joints = spec.config.joints

    @register.states()
    def reset(self):
        pass

    # TODO: fix correct limits
    @register.inputs(
        cartesian_pos=Space(shape=(len(go1_config.NOMINAL_FOOT_POS_LEG_FRAME),), dtype="float32"),  # TODO: Set correct bounds
    )
    @register.outputs(joint_pos=Space(low=go1_config.RL_LOWER_ANGLE_JOINT, high=go1_config.RL_UPPER_ANGLE_JOINT))
    def callback(self, t_n: float, cartesian_pos: Msg):
        # desired [x, y, z] for each joint
        action = np.array(cartesian_pos.msgs[-1].data)
        desired_joint_angles = np.zeros(len(self.joints))
        # call inverse kinematics to get corresponding joint angles
        for leg_idx in range(go1_config.NUM_LEGS):
            xyz_desired = action[3 * leg_idx : 3 * (leg_idx + 1)]
            leg_q = self.compute_inverse_kinematics(leg_idx, xyz_desired)
            desired_joint_angles[3 * leg_idx : 3 * (leg_idx + 1)] = leg_q

        # Send desired joint positions.
        return dict(joint_pos=desired_joint_angles.astype("float32"))

    @staticmethod
    def compute_inverse_kinematics(leg_id: int, xyz_coord: np.ndarray):
        """Get joint angles for leg leg_id with desired xyz position in leg frame.

        Leg 0: FR; Leg 1: FL; Leg 2: RR ; Leg 3: RL;

        From SpotMicro:
        https://github.com/OpenQuadruped/spot_mini_mini/blob/spot/spotmicro/Kinematics/LegKinematics.py
        """
        # rename links
        shoulder_length = go1_config.HIP_LINK_LENGTH
        elbow_length = go1_config.THIGH_LINK_LENGTH
        wrist_length = go1_config.CALF_LINK_LENGTH
        # coords
        x = xyz_coord[0]
        y = xyz_coord[1]
        z = xyz_coord[2]

        # get_domain
        D = (y**2 + (-z) ** 2 - shoulder_length**2 + (-x) ** 2 - elbow_length**2 - wrist_length**2) / (
            2 * wrist_length * elbow_length
        )

        D = np.clip(D, -1.0, 1.0)

        # check Right vs Left leg for hip angle
        sideSign = 1
        if leg_id == 0 or leg_id == 2:
            sideSign = -1

        # Right Leg Inverse Kinematics Solver
        wrist_angle = np.arctan2(-np.sqrt(1 - D**2), D)
        sqrt_component = y**2 + (-z) ** 2 - shoulder_length**2
        if sqrt_component < 0.0:
            sqrt_component = 0.0
        shoulder_angle = -np.arctan2(z, y) - np.arctan2(np.sqrt(sqrt_component), sideSign * shoulder_length)
        elbow_angle = np.arctan2(-x, np.sqrt(sqrt_component)) - np.arctan2(
            wrist_length * np.sin(wrist_angle), elbow_length + wrist_length * np.cos(wrist_angle)
        )
        joint_angles = np.array([-shoulder_angle, elbow_angle, wrist_angle])
        return joint_angles



import os
import site
import cv2
import subprocess
import base64
import numpy as np
from tqdm import tqdm
from datetime import datetime
from stable_baselines3.common.vec_env import VecVideoRecorder, DummyVecEnv
from pathlib import Path
from IPython import display as ipythondisplay


def setup_notebook():
    if "google.colab" in str(get_ipython()):  # noqa:
        print("Running on CoLab.")
        # Required to source ros in new processes
        os.environ["EAGERX_COLAB"] = "1"
    else:
        print("Not running on CoLab.")

    os.environ["EAGERX_RELOAD"] = "1"


def deprecated_setup_notebook():
    if "google.colab" in str(get_ipython()):  # noqa:
        print("Running on CoLab.")
        # Required to source ros in new processes
        os.environ["EAGERX_COLAB"] = "1"
        # Set paths to ROS libraries (instead of sourcing)
        site.addsitedir("/opt/ros/melodic/lib/python2.7/dist-packages")
        site.addsitedir("/usr/lib/python2.7/dist-packages")
        os.environ["ROS_PACKAGE_PATH"] = "/opt/ros/melodic/share/"
    else:
        print("Not running on CoLab.")
        print('Execute ROS commands as "!...".')

    os.environ["EAGERX_RELOAD"] = "1"

    try:
        import rospy, roslaunch, rosparam, rospkg  # noqa:

        pkg = rospkg.RosStack()
        if "noetic" in pkg.get_path("ros"):
            ros_version = "noetic"
        elif "melodic" in pkg.get_path("ros"):
            ros_version = "melodic"
        else:
            raise ModuleNotFoundError("No ROS version (noetic, melodic) available. Check if installed & sourced.")
        print(f"ROS {ros_version} available.")
    except ModuleNotFoundError:
        ModuleNotFoundError("No ROS version (noetic, melodic) available. Check if installed & sourced.")

    # No restart possible.
    if "google.colab" in str(get_ipython()):  # noqa:
        timestamp = datetime.today().strftime("%H:%M:%S")
        try:
            roscore.terminate()
            print(f"Roscore restarted: {timestamp}.")
        except NameError:
            print(f"Roscore started: {timestamp}.")

        # Start the roscore node
        roscore = subprocess.Popen(["bash", "/root/roscore"])  # noqa:


def record_video(env, model, video_length=500, prefix="", video_folder="videos/"):
    """
    Adapted from https://colab.research.google.com/github/Stable-Baselines-Team/rl-colab-notebooks/blob/sb3/stable_baselines_getting_started.ipynb

    :param env_id: (str)
    :param model: (RL model)
    :param video_length: (int)
    :param prefix: (str)
    :param video_folder: (str)
    """
    eval_env = DummyVecEnv([lambda: env])

    # Start the video at step=0 and record for video length
    eval_env = VecVideoRecorder(
        eval_env,
        video_folder=video_folder,
        record_video_trigger=lambda step: step == 0,
        video_length=video_length,
        name_prefix=prefix,
    )

    obs = eval_env.reset()
    for _ in range(video_length):
        action, _ = model.predict(obs, deterministic=True)
        obs, _, _, _ = eval_env.step(action)

    # Close the video recorder
    eval_env.close()


def show_video(video_file, video_folder="videos/"):
    """
    Adapted from https://colab.research.google.com/github/Stable-Baselines-Team/rl-colab-notebooks/blob/sb3/stable_baselines_getting_started.ipynb

    which was taken from https://github.com/eleurent/highway-env

    :param video_path: (str) Path to the folder containing videos
    :param prefix: (str) Filter the video, showing only the only starting with this prefix
    """
    html = []
    mp4 = Path(video_folder) / f"{video_file}.mp4"
    video_b64 = base64.b64encode(mp4.read_bytes())
    html.append(
        """<video alt="{}" autoplay 
                loop controls style="height: 400px;">
                <source src="data:video/mp4;base64,{}" type="video/mp4" />
            </video>""".format(
            mp4, video_b64.decode("ascii")
        )
    )
    ipythondisplay.display(ipythondisplay.HTML(data="<br>".join(html)))


def evaluate(model, env, n_eval_episodes=3, episode_length=100, video_rate=None, video_prefix=""):
    video_folder = "videos/"
    # Create output folder if needed
    os.makedirs(video_folder, exist_ok=True)

    episodic_rewards = []
    for i in range(n_eval_episodes):
        print(f"Start evaluation episode {i} of {n_eval_episodes}")
        img_array = []
        episodic_reward = 0
        obs = env.reset()
        for _step in tqdm(range(episode_length)):
            action, _ = model.predict(obs, deterministic=True)
            obs, reward, done, info = env.step(action)
            episodic_reward += reward
            if video_rate is not None:
                img = env.render("rgb_array")
                if 0 not in img.shape:
                    img_array.append(img)

        if video_rate is not None:
            video_file = f"{video_prefix}_{i}"
            path = Path(video_folder) / video_file
            print("Start video writer")
            height, width, _ = img_array[-1].shape
            size = (width, height)
            fourcc = cv2.VideoWriter_fourcc(*"mp4v")
            out = cv2.VideoWriter(f"{video_folder}/temp.mp4", fourcc, video_rate, size)

            for img in img_array:
                img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
                out.write(img)
            out.release()

            os.system(
                f"ffmpeg -y -i {video_folder}/temp.mp4 -vcodec libx264 -f mp4 {path}.mp4 >> /tmp/ffmpeg_{video_prefix}_{i}.txt 2>&1"
            )

            print(f"Showing episode {i} with episodic reward: {episodic_reward}")
            show_video(video_file=video_file, video_folder=video_folder)

            os.remove(f"{video_folder}/temp.mp4")

        episodic_rewards.append(episodic_reward)
    mean_episodic_reward = np.mean(episodic_rewards)
    print(f"Finished evaluation with mean episodic reward: {mean_episodic_reward}")



import eagerx
import numpy as np
from eagerx.core.specs import ProcessorSpec


class DecomposedAngle(eagerx.Processor):
    @classmethod
    def make(cls, convert_to: str = "theta_dtheta") -> ProcessorSpec:
        spec = cls.get_specification()
        return spec

    def initialize(self, spec: ProcessorSpec):
        pass

    def convert(self, msg: np.ndarray) -> np.ndarray:
        return np.array([np.cos(msg), np.sin(msg)], dtype="float32")


class ObsWithDecomposedAngle(eagerx.Processor):
    @classmethod
    def make(cls, convert_to: str = "theta_dtheta") -> ProcessorSpec:
        spec = cls.get_specification()
        spec.config.convert_to = convert_to
        return spec

    def initialize(self, spec: ProcessorSpec):
        self.convert_to = spec.config.convert_to

    def convert(self, msg: np.ndarray) -> np.ndarray:
        if not len(msg):  # No data
            data = np.array(msg, dtype="float32")
        elif self.convert_to == "trig_dtheta":
            data = np.array([np.sin(-msg.data[0]), np.cos(msg.data[0]), -msg.data[1]], dtype="float32")
        elif self.convert_to == "theta_dtheta":
            cos_th = msg.data[0]
            sin_th = msg.data[1]
            data = np.array([-np.arctan2(sin_th, cos_th), -msg.data[2]], dtype="float32")
        else:
            raise NotImplementedError(f"Convert_to '{self.convert_to}' not implemented.")
        return data


class Negate(eagerx.Processor):
    @classmethod
    def make(cls) -> ProcessorSpec:
        return cls.get_specification()

    def initialize(self, spec: ProcessorSpec):
        pass

    def convert(self, msg: np.ndarray) -> np.ndarray:
        return -msg


class VoltageToMotorTorque(eagerx.Processor):
    @classmethod
    def make(cls, K: float, R: float) -> ProcessorSpec:
        # Initialize spec with default arguments
        spec = cls.get_specification()
        spec.config.K = K
        spec.config.R = R
        return spec

    def initialize(self, spec: ProcessorSpec):
        self.K = spec.config.K
        self.R = spec.config.R

    def convert(self, msg: np.ndarray) -> np.ndarray:
        return np.array([-msg[0] * self.K / self.R], dtype="float32")



from matplotlib import animation
import matplotlib.pyplot as plt


def save_frames_as_gif(dt, frames, path=".", filename="swimm_animation.gif", dpi=15):
    # Mess with this to change frame size
    fig = plt.figure(figsize=(frames[0].shape[1] / 72, frames[0].shape[0] / 72.0), dpi=dpi)
    ax = fig.gca()
    ax.set_axis_off()
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
    plt.margins(0, 0)
    ax.xaxis.set_major_locator(plt.NullLocator())
    ax.yaxis.set_major_locator(plt.NullLocator())

    patch = plt.imshow(frames[0])
    plt.axis("off")

    def animate(i):
        patch.set_data(frames[i])

    anim = animation.FuncAnimation(plt.gcf(), animate, frames=len(frames), interval=50)
    anim.save("%s/%s" % (path, filename), writer="Pillow", fps=int(1 / dt))
    plt.close(fig)
    print("Gif saved to %s/%s" % (path, filename))



from typing import Optional, List
import numpy as np
import eagerx
from eagerx import Space
from eagerx.utils.utils import Msg


def wrap_angle(angle):
    return angle - 2 * np.pi * np.floor((angle + np.pi) / (2 * np.pi))


class ResetAngle(eagerx.ResetNode):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        threshold: float = 0.1,
        timeout: float = 5.0,
        gains: Optional[List[float]] = None,
        u_range: Optional[List[float]] = None,
    ):
        """This AngleReset node resets the pendulum to a desired angle with zero angular velocity. NOte that this controller
        only works properly when resetting the pendulum near the downward facing equilibrium.

        :param name: Node name
        :param rate: Rate at which callback is called.
        :param threshold: Absolute difference between the desired and goal state, before considering the reset complete.
        :param timeout: Maximum time (seconds) before considering the reset finished (regardless whether the goal was reached).
        :param gains: Gains of the PID controller used to reset.
        :param u_range: Min and max action.
        :return:
        """
        spec = cls.get_specification()

        # Modify default node params
        spec.config.update(name=name, rate=rate, process=eagerx.ENVIRONMENT, color="grey")
        spec.config.update(inputs=["theta", "dtheta"], targets=["goal"], outputs=["u"])
        spec.config.update(u_range=u_range, threshold=threshold, timeout=timeout)
        spec.config.gains = gains if isinstance(gains, list) else [1.0, 0.5, 0.0]

        # Add space_converter
        spec.outputs.u.space = Space(low=[u_range[0]], high=[u_range[1]], dtype="float32")
        return spec

    def initialize(self, spec):
        self.threshold = spec.config.threshold
        self.timeout = spec.config.timeout
        self.u_min, self.u_max = spec.config.u_range

        # Creat a simple PID controller
        from eagerx_tutorials.pendulum.pid import PID

        gains = spec.config.gains
        self.controller = PID(u0=0.0, kp=gains[0], kd=gains[1], ki=gains[2], dt=1 / self.rate)

    @eagerx.register.states()
    def reset(self):
        # Reset the internal state of the PID controller (ie the error term).
        self.controller.reset()
        self.ts_start_routine = None

    @eagerx.register.inputs(
        theta=Space(shape=(), dtype="float32"),
        dtheta=Space(shape=(), dtype="float32"),
    )
    @eagerx.register.targets(goal=Space(low=[-3.14, -9.0], high=[3.14, 9.0]))
    @eagerx.register.outputs(u=Space(low=[-2.0], high=[2.0]))
    def callback(self, t_n: float, goal: Msg, theta: Msg = None, dtheta: Msg = None, x: Msg = None):
        if self.ts_start_routine is None:
            self.ts_start_routine = t_n

        # Convert messages to floats and numpy array
        theta = theta.msgs[-1]  # Take the last received message
        dtheta = dtheta.msgs[-1]  # Take the last received message
        goal = np.array(goal.msgs[-1], dtype="float32")  # Take the last received message

        # Define downward angle as theta=0 (resolve downward discontinuity)
        theta += np.pi
        goal[0] += np.pi

        # Wrap angle between [-pi, pi]
        theta = wrap_angle(theta)
        goal[0] = wrap_angle(goal[0])

        # Overwrite the desired velocity to be zero.
        goal[1] = 0.0

        # Calculate the action using the PID controller
        u = self.controller.next_action(theta, ref=goal[0])
        u = np.clip(u, self.u_min, self.u_max)  # Clip u to range

        # Determine if we have reached our goal state
        done = np.isclose(np.array([theta, dtheta]), goal, atol=self.threshold).all()

        # If the reset routine takes too long, we timeout the routine and simply assume that we are done.
        done = done.item() or (t_n - self.ts_start_routine) > self.timeout

        # Prepare output message for transmission.
        # This must contain a message for every registered & selected output and target.
        # For targets, this message decides whether the goal state has been reached (or we, for example, timeout the reset).
        # The name for this target message is the registered target name + "/done".
        output_msgs = {"u": np.array([u], dtype="float32"), "goal/done": done}
        return output_msgs



import os
from typing import List, Optional

import eagerx
from eagerx import Space
import eagerx.core.register as register
from eagerx.core.graph_engine import EngineGraph
from eagerx.core.specs import ObjectSpec
from eagerx_pybullet.engine import PybulletEngine

import eagerx_tutorials.quadruped.go1.configs_go1 as go1_config


class Quadruped(eagerx.Object):
    @classmethod
    @register.sensors(
        joint_position=Space(low=go1_config.RL_LOWER_ANGLE_JOINT, high=go1_config.RL_UPPER_ANGLE_JOINT, dtype="float32"),
        joint_velocity=Space(low=-go1_config.VELOCITY_LIMITS, high=go1_config.VELOCITY_LIMITS, dtype="float32"),
        force_torque=Space(  # todo: set realistic bounds for forces and moments.
            low=[-200.0] * 6 * len(go1_config.RL_LOWER_ANGLE_JOINT),
            high=[200.0] * 6 * len(go1_config.RL_LOWER_ANGLE_JOINT),
            dtype="float32",
        ),
        orientation=Space(low=go1_config.INIT_ORIENTATION, high=go1_config.INIT_ORIENTATION, dtype="float32"),
        position=Space(low=[-10.0, -10.0, 0.0], high=[10.0, 10.0, 0.5], dtype="float32"),
        velocity=Space(low=[-1.0, -1.0, -0.2], high=[1.0, 1.0, 0.2], dtype="float32"),
        image=Space(dtype="uint8"),
    )
    @register.actuators(
        joint_control=Space(low=go1_config.RL_LOWER_ANGLE_JOINT, high=go1_config.RL_UPPER_ANGLE_JOINT),
    )
    @register.engine_states(
        joint_position=Space(low=go1_config.INIT_JOINT_ANGLES, high=go1_config.INIT_JOINT_ANGLES, dtype="float32"),
        position=Space(low=go1_config.INIT_POSITION, high=go1_config.INIT_POSITION, dtype="float32"),
        orientation=Space(low=go1_config.INIT_ORIENTATION, high=go1_config.INIT_ORIENTATION, dtype="float32"),
        velocity=Space(low=[0.0, 0.0, 0.0], high=[0.0, 0.0, 0.0], dtype="float32"),
        angular_velocity=Space(low=[0.0, 0.0, 0.0], high=[0.0, 0.0, 0.0], dtype="float32"),
    )
    def make(
        cls,
        name: str,
        sensors: Optional[List[str]] = None,
        actuators: Optional[List[str]] = None,
        states: Optional[List[str]] = None,
        rate: float = 30.0,
        position: Optional[List[int]] = None,
        orientation: Optional[List[int]] = None,
        self_collision: bool = False,
        fixed_base: bool = False,
        control_mode: str = "position_control",
        render_shape: Optional[List[int]] = None,
    ) -> ObjectSpec:
        """Make a spec to create a go1 quadruped robot.

        :param name: Name of the object (topics are placed within this namespace).
        :param sensors: A list of selected sensors. Must be a subset of the registered sensors.
        :param actuators: A list of selected actuators. Must be a subset of the registered actuators.
        :param states: A list of selected states. Must be a subset of the registered actuators.
        :param rate: The default rate at which all sensors and actuators run. Can be modified via the spec API.
        :param position: Base position of the object [x, y, z].
        :param orientation: Base orientation of the object in quaternion [x, y, z, w].
        :param self_collision: Enable self collisions.
        :param fixed_base: Force the base of the loaded object to be static.
        :param control_mode: Control mode for the arm joints. Available: `position_control`, `velocity_control`, `pd_control`, and `torque_control`.
        :param render_shape: Render shape (height, width) of the image produced by the image sensor.
        :return: ObjectSpec
        """
        spec = cls.get_specification()

        # Modify default agnostic params
        # Only allow changes to the agnostic params (rates, windows, (space)converters, etc...
        spec.config.name = name
        spec.config.sensors = sensors if sensors else ["joint_position", "joint_velocity", "orientation"]
        spec.config.actuators = actuators if actuators else ["joint_control"]
        spec.config.states = (
            states if states else ["joint_position", "position", "orientation", "velocity", "angular_velocity"]
        )

        # Add registered agnostic params
        spec.config.joint_names = list(go1_config.JOINT_NAMES)
        spec.config.position = position if position else go1_config.INIT_POSITION
        spec.config.orientation = orientation if orientation else [0, 0, 0, 1]
        spec.config.self_collision = self_collision
        spec.config.fixed_base = fixed_base
        spec.config.control_mode = control_mode
        spec.config.render_shape = render_shape if isinstance(render_shape, list) else [400, 400]

        # Set rates
        spec.sensors.joint_position.rate = rate
        spec.sensors.joint_velocity.rate = rate
        spec.sensors.force_torque.rate = rate
        spec.sensors.orientation.rate = rate
        spec.sensors.position.rate = rate
        spec.sensors.velocity.rate = rate
        spec.sensors.image.rate = rate
        spec.actuators.joint_control.rate = rate

        # Set variable space limits
        spec.sensors.image.space = Space(
            dtype="uint8",
            low=0,
            high=255,
            shape=tuple(spec.config.render_shape + [3]),
        )

        return spec

    @staticmethod
    @register.engine(PybulletEngine)
    def pybullet_engine(spec: ObjectSpec, graph: EngineGraph):
        """Engine-specific implementation (Pybullet) of the object."""
        # Set object arguments (as registered per register.engine_params(..) above the engine.add_object(...) method.
        urdf_file = os.path.join(go1_config.URDF_ROOT, go1_config.URDF_FILENAME)
        spec.engine.urdf = urdf_file
        spec.engine.basePosition = spec.config.position
        spec.engine.baseOrientation = spec.config.orientation
        spec.engine.fixed_base = spec.config.fixed_base
        spec.engine.self_collision = spec.config.self_collision

        # Create engine_states (no agnostic states defined in this case)
        from eagerx_pybullet.enginestates import JointState, LinkState

        spec.engine.states.joint_position = JointState.make(joints=spec.config.joint_names, mode="position")

        spec.engine.states.position = LinkState.make(mode="position")
        spec.engine.states.orientation = LinkState.make(mode="orientation")
        spec.engine.states.velocity = LinkState.make(mode="velocity")
        spec.engine.states.angular_velocity = LinkState.make(mode="angular_vel")

        # Create sensor engine nodes
        # Rate=None, but we will connect them to sensors (thus will use the rate set in the agnostic specification)
        from eagerx_pybullet.enginenodes import JointSensor, LinkSensor, CameraSensor, JointController

        rate = spec.sensors.joint_position.rate
        joint_position = JointSensor.make(
            "joint_position",
            rate=rate,
            process=eagerx.process.ENGINE,
            joints=spec.config.joint_names,
            mode="position",
        )

        joint_velocity = JointSensor.make(
            "joint_velocity",
            rate=rate,
            process=eagerx.process.ENGINE,
            joints=spec.config.joint_names,
            mode="velocity",
        )
        force_torque = JointSensor.make(
            "force_torque",
            rate=rate,
            process=eagerx.process.ENGINE,
            joints=spec.config.joint_names,
            mode="force_torque",
        )

        # TODO: convert to euler (currently quaternion)
        orientation = LinkSensor.make(
            "orientation",
            rate=rate,
            process=eagerx.process.ENGINE,
            links=None,
            mode="orientation",
        )
        position = LinkSensor.make(
            "position",
            rate=rate,
            process=eagerx.process.ENGINE,
            links=None,
            mode="position",
        )
        velocity = LinkSensor.make(
            "velocity",
            rate=rate,
            process=eagerx.process.ENGINE,
            links=None,
            mode="velocity",
        )
        image = CameraSensor.make(
            "image",
            rate=spec.sensors.image.rate,
            process=eagerx.process.ENGINE,
            mode="rgb",
            render_shape=spec.config.render_shape,
        )
        image.config.fov = 80.0  # todo: tune

        # Create actuator engine nodes
        # Rate=None, but we will connect it to an actuator (thus will use the rate set in the agnostic specification)
        joint_control = JointController.make(
            "joint_control",
            rate=spec.actuators.joint_control.rate,
            process=eagerx.process.ENGINE,
            joints=spec.config.joint_names,
            mode=spec.config.control_mode,
            vel_target=[0] * len(go1_config.JOINT_NAMES),
            pos_gain=[1] * len(go1_config.JOINT_NAMES),
            vel_gain=[1] * len(go1_config.JOINT_NAMES),
            max_force=[10] * len(go1_config.JOINT_NAMES),
        )
        # Connect all engine nodes
        graph.add(
            [
                joint_position,
                joint_velocity,
                force_torque,
                orientation,
                position,
                velocity,
                joint_control,
                image,
            ]
        )
        graph.connect(source=joint_position.outputs.obs, sensor="joint_position")
        graph.connect(source=joint_velocity.outputs.obs, sensor="joint_velocity")
        graph.connect(source=force_torque.outputs.obs, sensor="force_torque")
        graph.connect(source=orientation.outputs.obs, sensor="orientation")
        graph.connect(source=position.outputs.obs, sensor="position")
        graph.connect(source=velocity.outputs.obs, sensor="velocity")
        graph.connect(source=image.outputs.image, sensor="image")
        graph.connect(actuator="joint_control", target=joint_control.inputs.action)



"""Defines the GO1 robot related constants and URDF specs."""
import inspect
import os
import re

import numpy as np
import pybullet as pyb

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)

URDF_ROOT = parentdir
URDF_FILENAME = "go1/go1_description/urdf/go1.urdf"

##################################################################################
# Default robot configuration (i.e. base and joint positions, etc.)
##################################################################################
NUM_MOTORS = 12
NUM_LEGS = 4
MOTORS_PER_LEG = 3

INIT_RACK_POSITION = [0, 0, 1]  # when hung up in air (for debugging)
INIT_POSITION = [0, 0, 0.33]  # normal initial height
IS_FALLEN_HEIGHT = 0.18  # height at which robot is considered fallen

INIT_ORIENTATION = (0, 0, 0, 1)
_, INIT_ORIENTATION_INV = pyb.invertTransform(position=[0, 0, 0], orientation=INIT_ORIENTATION)

# default angles (for init)
DEFAULT_HIP_ANGLE = 0  # anca
DEFAULT_THIGH_ANGLE = np.pi / 4  # coscia
DEFAULT_CALF_ANGLE = -np.pi / 2  # ginocchio

INIT_JOINT_ANGLES = np.array([DEFAULT_HIP_ANGLE, DEFAULT_THIGH_ANGLE, DEFAULT_CALF_ANGLE] * NUM_LEGS, dtype="float32")
INIT_MOTOR_ANGLES = INIT_JOINT_ANGLES
# Used to convert the robot SDK joint angles to URDF joint angles.
JOINT_DIRECTIONS = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

# joint offsets
HIP_JOINT_OFFSET = 0.0
THIGH_JOINT_OFFSET = 0.0
CALF_JOINT_OFFSET = 0.0

# Used to convert the robot SDK joint angles to URDF joint angles.
JOINT_OFFSETS = np.array([HIP_JOINT_OFFSET, THIGH_JOINT_OFFSET, CALF_JOINT_OFFSET] * NUM_LEGS, dtype="float32")

# Kinematics
HIP_LINK_LENGTH = 0.0847
THIGH_LINK_LENGTH = 0.213
CALF_LINK_LENGTH = 0.213

NOMINAL_FOOT_POS_LEG_FRAME = np.array(
    [0, -HIP_LINK_LENGTH, -0.32, 0, HIP_LINK_LENGTH, -0.32, 0, -HIP_LINK_LENGTH, -0.32, 0, HIP_LINK_LENGTH, -0.32],
    dtype="float32",
)

##################################################################################
# Actuation limits/gains, position, and velocity limits
##################################################################################
# joint limits on real system
REAL_UPPER_ANGLE_JOINT = np.array([1.0471975512, 2.96705972839, -0.837758040957] * NUM_LEGS, dtype="float32")
REAL_LOWER_ANGLE_JOINT = np.array([-1.0471975512, -0.663225115758, -2.72271363311] * NUM_LEGS, dtype="float32")

# modified range in simulation (min observation space for RL)
RL_UPPER_ANGLE_JOINT = np.array([0.2, DEFAULT_THIGH_ANGLE + 0.6, DEFAULT_CALF_ANGLE + 0.6] * NUM_LEGS, dtype="float32")
RL_LOWER_ANGLE_JOINT = np.array([-0.2, DEFAULT_THIGH_ANGLE - 0.6, DEFAULT_CALF_ANGLE - 0.6] * NUM_LEGS, dtype="float32")

# torque and velocity limits
# Set to 0.4 * ... to limit max torque
TORQUE_LIMITS = 1.0 * np.asarray([23.7, 23.7, 1.0 * 33.55] * NUM_LEGS, dtype="float32")
VELOCITY_LIMITS = 1.0 * np.asarray([30.1, 30.1, 30.1] * NUM_LEGS, dtype="float32")

# Sample Joint Gains
MOTOR_KP = [100.0, 100.0, 100.0] * NUM_LEGS
MOTOR_KD = [1.0, 2.0, 2.0] * NUM_LEGS

MOTOR_KP = [55, 55, 55] * NUM_LEGS
MOTOR_KD = [0.8, 0.8, 0.8] * NUM_LEGS

# Sample Cartesian Gains
kpCartesian = np.diag([500, 500, 500])
kdCartesian = np.diag([10, 10, 10])

# kpCartesian = np.diag([700, 700, 700])
# kdCartesian = np.diag([12, 12, 12])

##################################################################################
# Hip, thigh, calf strings, naming conventions from URDF (don't modify)
##################################################################################
JOINT_NAMES = (
    # front right leg
    "FR_hip_joint",
    "FR_thigh_joint",
    "FR_calf_joint",
    # front left leg
    "FL_hip_joint",
    "FL_thigh_joint",
    "FL_calf_joint",
    # rear right leg
    "RR_hip_joint",
    "RR_thigh_joint",
    "RR_calf_joint",
    # rear left leg
    "RL_hip_joint",
    "RL_thigh_joint",
    "RL_calf_joint",
)
MOTOR_NAMES = JOINT_NAMES

# standard across all robots
_CHASSIS_NAME_PATTERN = re.compile(r"\w*floating_base\w*")
_HIP_NAME_PATTERN = re.compile(r"\w+_hip_j\w+")
_THIGH_NAME_PATTERN = re.compile(r"\w+_thigh_j\w+")
_CALF_NAME_PATTERN = re.compile(r"\w+_calf_j\w+")
_FOOT_NAME_PATTERN = re.compile(r"\w+_foot_\w+")

############################################
# Spring level joint Variables
############################################
# stiffness
K_HIP = 25
K_THIGH = 25
K_CALF = 30
# damping
D_HIP = 0.4
D_TIHGH = 0.4
D_CALF = 0.4

SPRINGS_STIFFNESS = [K_HIP, K_CALF, K_THIGH]
SPRINGS_DAMPING = [D_HIP, D_TIHGH, D_CALF]
SPRINGS_REST_ANGLE = [DEFAULT_HIP_ANGLE, DEFAULT_THIGH_ANGLE, DEFAULT_CALF_ANGLE + 0.3]

################################################
# Parameters for actions clipping
################################################
# step size is equal to 0.001 * action_repeat(default_value = 10)
MAX_MOTOR_ANGLE_CHANGE_PER_STEP = 0.2
MAX_CARTESIAN_FOOT_POS_CHANGE_PER_STEP = np.array([0.1, 0.02, 0.08], dtype="float32")
